# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3. 
# See LICENSE comming with the source of 'mpuls offline'
# for details.
#
# author: Sascha L. Teichmann <sascha.teichmann@intevation.de>
#

from xml.sax          import make_parser
from xml.sax.handler  import ContentHandler
from xml.sax.saxutils import escape

from datetime         import date
from traceback        import print_exc

import codecs
import sys
import data 
import re
import shutil

from uuid import uuid4

from semantic import SematicError, checkAndConvert, ErrorItem

from data import PageNode, RuleLeaf, WidgetCollector

from navigation import Navigation

from string import printable

import os.path

# version of the loaded XML file
VERSION = 1

NO_NUMBER  = re.compile("[^0-9\-]")
INT_NUMBER = re.compile("(-?[0-9]+)")
DATE       = re.compile("([0-9]{1,4})-([0-9]{1,2})-([0-9]{1,2})")

WARNING    = re.compile(r"warning\s*:\s*([^\s]+)")

def decode_int(s):
    s = NO_NUMBER.sub("", s)
    m = INT_NUMBER.match(s)
    if not m: return None
    return int(m.group(1))

def decode_date(s):
    s = NO_NUMBER.sub("", s)
    m = DATE.match(s)
    if not m: return None
    return date(int(m.group(1)), int(m.group(2)), int(m.group(3)))

DECODERS = {
    data.RadioNode:    decode_int,
    data.TextAreaLeaf: lambda s: s,
    data.TextLeaf:     lambda s: s,
    data.ChoiceNode:   decode_int,
    data.IntLeaf:      decode_int,
    data.DateLeaf:     decode_date
}

ENCODERS = {
    date     : lambda x: "%d-%d-%d" % (x.year, x.month, x.day),
    str      : lambda x: escape(x),
    type(u''): lambda x: escape(x)
}

XML_HEADER = \
u'''<?xml version="1.0" encoding="UTF-8"?>
<cases version="%s">
<case>
<master>'''

XML_FOOTER = \
u'''</master>
</case>
</cases>'''

class ElementBuilder(ContentHandler):

    def __init__(self, start, end):
        self.start = start
        self.end   = end
        self.stack = None

    def startElement(self, name, attrs):
        self.start(name, attrs)
        self.stack.append((attrs, []))

    def endElement(self, name):
        t = self.stack.pop()
        self.end(name, t[0], u"".join(t[1]))

    def characters(self, chars):
        self.stack[-1][1].append(chars)

    def startDocument(self):
        self.stack = []
    
class MemoryDatastore(object):

    def __init__(self, formed, defaults, filename = None):
        self.formed   = formed
        self.data     = {}
        self.dirty    = False
        self.defaults = defaults

        collector = WidgetCollector()
        formed.visit(collector.visitor)
        self.widgets = dict([(w.getName(), w) for w in collector.widgets if w.getName()])

        if filename:
            self.load(filename)
        else:
            self.presetDefaults()

        self.filename = filename
        self.swapfile = None
        self.errors   = {}
        self.navigation = None

    def setSwapfile(self, swapfile=None):
        self.swapfile = swapfile

    def presetDefaults(self):
        defaults = self.defaults
        for item in defaults.values:
            try:
                w = self.widgets[item.key]
            except KeyError:
                continue
            try:
                iv = item.eval()
            except:
                print_exc()
                continue
            decoder = DECODERS.get(w.__class__, lambda s: s)
            try:
                iv = decoder(iv)
                if iv and w.findDefault() != iv:
                    self.data[item.key] = iv
            except:
                print_exc()

    def isDirty(self):
        return self.dirty

    def getData(self, name):
        try:
            return self.data[name]
        except KeyError:
            try:
                value = self.widgets[name].findDefault()
                self.data[name] = value
                return value
            except KeyError:
                return None

    def getFilename(self):
        return self.filename

    def getErrors(self):
        return self.errors

    def getPageNamesWithErrors(self):
        return set([e.page for e in self.errors.itervalues()])

    def getNavigation(self):
        if self.navigation is None:
            self.navigation = Navigation(self.formed)
        return self.navigation

    def deleteData(self, name):
        try:
            v = self.data[name]
            d = self.widgets[name].findDefault()
            if v != d:
                self.data[name] = d
                self.dirty = True
                return True
            return False
        except KeyError:
            return False

    def setData(self, name, value):
        try:
            old = self.data[name]
            if old != value:
                self.data[name] = value
                self.dirty = True
                return True
            return False
        except KeyError:
            w = self.widgets.get(name)
            if w and w.findDefault() != value:
                self.data[name] = value
                self.dirty = True
                return True
        return False

    def setFilename(self, filename):
        if self.filename != filename:
            self.filename = filename
            self.dirty = True

    def load(self, filename):

        widgets = self.widgets

        new_data = {}

        def element_start(name, attrs):
            if name == u"cases":
                try:
                    version = int(attrs[u'version'])
                except (KeyError, ValueError):
                    version = 1

                if version > VERSION:
                    raise Exception(
                        "Cannot load files of version %d." % version)

        def element_end(name, attrs, value):

            try:
                widget = widgets[name]
            except KeyError:
                if name == u'uuid_id':
                    new_data[name] = value
                #print >> sys.stderr, "Tag not found: %s" % repr(name)
                return

            new_data[name] = DECODERS.get(
                widget.__class__, lambda s: s)(value)

        f = None
        try:
            f = open(filename.encode(sys.getfilesystemencoding()), "rb")
            parser = make_parser()
            parser.setContentHandler(ElementBuilder(
                element_start,
                element_end))
            parser.parse(f)
        finally:
            if f:
                try: f.close()
                except: pass

        self.filename = filename
        self.data     = new_data
        self.dirty    = False

    def store(self, filename=None, backup=True):
        if filename is None:
            if not self.filename: return
            filename = self.filename

        collector = WidgetCollector()
        self.formed.visit(collector.visitor)

        data = self.data

        # Write a backup file before storing data. The backup should always
        # contain the data of the last save
        if backup and os.path.isfile(filename):
            bak = filename + ".bak"
            try:
                shutil.copy(filename, bak)
            except OSError, e:
                pass

        out = None
        try:
            out = codecs.open(filename, "w", "UTF-8")
            print >> out, XML_HEADER % VERSION

            try:
                uuid_id = data['uuid_id']
            except KeyError:
                uuid_id = str(uuid4())
                data['uuid_id'] = uuid_id

            print >> out, '<uuid_id>%s</uuid_id>' % uuid_id

            for widget in collector.widgets:
                name = widget.getName()
                try:
                    value = data[name]
                    if value is None: continue
                except KeyError:
                    continue
                encoder = ENCODERS.get(type(value), lambda x: u"%s" % x)
                print >> out, u"<%s>%s</%s>" % (name, encoder(value), name)
            print >> out, XML_FOOTER

            if backup: self.dirty = False
        finally:
            if out:
                try: out.close()
                except: pass

    def warningsOnPage(self, pageName):
        """ Generates a dictionary of warnings that should be displayed
            on a given page.
        """
        # find all widgets on page
        document = self.formed
        nc = document.findByClassAndName(PageNode, pageName)
        if not nc: return {}

        # widgets on current page
        used_fields = set([
            w.getName() for w in nc.allWidgets() if w.getName()])

        warnings = {}

        for r in document.findAllByClass(RuleLeaf):
            mark, expr = r.getMark(), r.getExpr()
            if not mark or not expr: continue
            # warnings only
            mark = WARNING.search(mark)
            if not mark: continue
            mark = mark.group(1)
            # only mark if field is on current page
            if not mark in used_fields: continue

            try:
                okay = expr.evaluate(dict([
                    (n, self.getData(n)) for n in expr.getDependencies()]))
            except:
                print_exc()
                okay = False

            if not okay:
                try:
                    t = warnings[mark]
                except KeyError:
                    t = (self.widgets[mark], [])
                    warnings[mark] = t
                t[1].append(r.getValue())

        return warnings

    def save(self, params, pageName):

        document = self.formed
        nc = document.findByClassAndName(PageNode, pageName)
        if not nc: return

        # widgets on current page
        widgets = dict([
            (widget.getName(), widget) for widget in nc.allWidgets() if widget.getName()])

        widgets_clone = widgets.copy()

        dbPage = self
        old_errors = self.errors

        # empty parameters will nullify values
        delete_vars = []

        # for updating the errors
        #old_errors = session.setdefault("$ERRORS", {})
        # Note: old_errors is now passed as argument

        # vars that passed lexical and syntactic check
        to_be_set = {}

        for k, v in params.iteritems():
            if k.startswith("__"):
                continue
            try:
                value = v[0]
                widget = widgets.pop(k)
                #print "widget: %s"  % repr(widget)
                if value == "": # no value -> delete it
                    delete_vars.append(k)
                else:
                    # Hopefully this one causes no trouble any more
                    old_errors.pop(k, None)
                    # the empty array is for the rules
                    nv = checkAndConvert(widget, value)
                    ov = self.getData(k)
                    if nv != ov:
                        to_be_set[k] = (nv, [])
            except SematicError, inst:
                ei = ErrorItem(pageName, value, widget.getDescription())
                ei.addMessage(inst.value)
                # overwrite old errors
                old_errors[k] = ei
            except KeyError, inst:
                pass

        # flag to indicate if we should flush the page
        dirty = False

        # remove variable from dataset which are set to ""
        for wname in list(widgets.iterkeys()) + delete_vars:
            if dbPage.deleteData(wname):
                dirty = True
            # an unset value causes no problems.
            old_errors.pop(wname, None)

        # find all rules that depend on variables to be set.
        # build up a lookup var-name -> list of rules
        for r in document.findAllByClass(RuleLeaf):
            mark = r.getMark()
            # warning rules are evaluated at page rendering
            if mark and mark.find("warning:") >= 0:
                #print "Ignore warning rule '%s'" % r.getName()
                continue
            expr = r.getExpr()
            if not expr: continue
            for var in expr.getDependencies():
                try:
                    to_be_set[var][1].append(r)
                except KeyError:
                    pass

        # set vars into db page and check rules.
        for k, tup in to_be_set.iteritems():
            v, rules = tup[0], tup[1]
            hasNoProblems = True
            for rule in rules:
                vars = {}
                isNull = False # only process rule if all values are not null
                for dependency in rule.getExpr().getDependencies():
                    if dependency == k: # ignore because this comes from the web
                        continue
                    # XXX: Normally they are distributed over all kind of pages.
                    value = self.getData(dependency)
                    if value is None: isNull = True; break
                    vars[dependency] = value
                if isNull:
                    continue
                expr = rule.getExpr()
                # Now evaluate the rule to see if its violated
                try:
                    vars[k] = v # test it with new value from web
                    isOkay = expr.evaluate(vars)
                except:
                    print_exc()
                    isOkay = False
                if not isOkay:
                    hasNoProblems = False
                    ie = old_errors.get(k, None)
                    if ie is None:
                        try:
                            desc = widgets_clone[k].getDescription()
                        except KeyError:
                            desc = ""
                        ie = ErrorItem(pageName, v, desc)
                        old_errors[k] = ie
                    msg = rule.getValue()
                    if msg: ie.addMessage(msg)
                    else:   ie.addMessage("There is a problem with '%s'." % k)
            if hasNoProblems: # No new errors -> set it
                if dbPage.setData(k, v):
                    dirty = True
                # TODO: Solve transitive problems!
                # e.g. A was set. B contradicts A. That prevents B from
                # being set. -> B is marked was wrong input.
                # Now A is set to a value that does not contradicts B
                # any more. Therefore B is not an error input any more.
                # -> B can be applied. This scheme has to be applied 
                # recursively along the dependencies.
                old_errors.pop(k, None)

        if dirty: # flush the db page
            self.dirty = True

        # Before continuing write a swap file, but do not overwrite the backup
        try:
            self.store(self.swapfile, backup=False)
        except:
            pass
        return old_errors


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
