# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3. 
# See LICENSE comming with the source of 'mpuls offline'
# for details.
#
# author: Sascha L. Teichmann <sascha.teichmann@intevation.de>
#

from nodecomponents import Node, RootNode
from data           import PageNode

from cgi import escape

class TreeItem(object):

    def __init__(self, name, description, children = None, parent = None):
        self.is_open     = False
        self.name        = name
        self.description = description
        self.parent      = parent
        self.children    = children

    def is_child(self):
        return self.children is None

    def deepest_item(self):
        cur = self
        while cur.children:
            cur = cur.children[0]
        return cur

    def unfold(self):
        cur = self
        while cur:
            cur.is_open = True
            cur = cur.parent

    def render(self, page, out, get_image, error_lookup):
        out.append(u'<li class="formpage">')
        out.append(page == self.name and u'<span class="formpage_highlight">' or u'<span>')

        has_error = self.name in error_lookup
        if self.is_child():

            select_link = 'javascript:select_item(\'%s\');' % self.name

            out.append(u'<a href="%s" class="navigation">' % select_link)

            if has_error:
                out.append(get_image("images/icons/formular_page_blue_error.gif",
                    u' border="0"'
                    u' alt="Fehler auf dieser Seite"'
                    u' title="Seite öffnen - Diese Seite enthält Fehler."'))
            else:
                out.append(get_image("images/icons/formular_page_blue.gif",
                    u' border="0"'
                    u' alt="Formularseite" title="Seite öffnen"'))

            out.append(u'</a> <a href="%s" class="navigation">%s</a>\n' % (
                select_link, escape(self.description)))
        else:
            # when clicking on link select the first page in branch
            deepest = self.deepest_item()

            select_link = 'javascript:select_branch(\'%s\');' %  deepest.name

            if self.is_open or has_error:
                if not has_error:
                    out.append(u'<a href="javascript:close_branch(\'%s\', \'%s\');" class="navigation">' % (
                        self.name, page))

                out.append(get_image("images/icons/folder_closed_blue_2.gif",
                    u' border="0"'
                    u' title="Ordner%s"'
                    u' alt="%s"' % (
                        not has_error and "schliessen" or "",
                        has_error and u"offener" or u"geschlossener")))

                out.append(has_error and u' ' or u'</a>')

                out.append(u'<a href="%s" class="navigation">%s</a>' % (
                    select_link, escape(self.description)))

                out.append(u'<ul>\n')
                for child in self.children:
                    child.render(page, out, get_image, error_lookup)
                out.append(u'</ul>\n')
            else:
                out.append(u'<a href="javascript:open_branch(\'%s\', \'%s\');" class="navigation">' % (
                    self.name, page))

                out.append(get_image("images/icons/folder_closed_blue_2.gif",
                    u' border="0"'
                    u' alt="geschlossener Ordner" title="Ordner öffnen"'))
                    
                out.append(u'</a> ')

                out.append(u'<a href="%s" class="navigation">%s</a>' % (
                    select_link, escape(self.description)))

        out.append(u"</span></li>\n")

    def set_open(self, flag):
        self.is_open = flag

    def build_error_lookup(self, lookup, error_pages):
        mark = self.name in error_pages
        if self.children:
            for child in self.children:
                if child.build_error_lookup(lookup, error_pages):
                    mark = True
        if mark:
            lookup.add(self.name)
        return mark

    def get_description_chain(self, form_data):
        cur = self
        descriptions = []
        while cur:
            if cur.description:
                descriptions.insert(0, u"%s%s" % (
                    escape(cur.description), form_data.getHelp(cur.name)))
            cur = cur.parent
        return descriptions

    def renderHeader(self, form_data):
        out = [u'<div class="content_header"><ul>']
        for depth, desc in enumerate(self.get_description_chain(form_data)):
            depth += 1
            out.append(u'<li><h%d>%s</h%d></li>' % (depth, desc, depth))
        out.append(u'</ul></div>')
        return u"".join(out)

class NavigationTreeBuilder(object):

    def __init__(self, formed):
        self.pages = []
        self.root = self.build(formed, None)

    def build(self, nc, parent):

        if isinstance(nc, PageNode):
            self.pages.append(nc.getName())
            return TreeItem(nc.getName(), nc.getDescription(), parent = parent)

        if isinstance(nc, Node):
            children = []
            n = TreeItem(nc.getName(), nc.getDescription(), children, parent)
            for child in nc.children:
                children.append(self.build(child, n))
            return n

        return TreeItem("xxx", "Unknown Item", parent = parent)

class Navigation(object):

    def __init__(self, formed):

        builder = NavigationTreeBuilder(formed.getRoot())

        self.root = builder.root

        ps = builder.pages

        pages = {}

        for idx in xrange(len(ps)):
            prev = idx > 0 and ps[idx-1] or None
            next = idx < len(ps)-1 and ps[idx+1] or None
            pages[ps[idx]] = (prev, next)

        self.pages = pages

    def render(self, page, file_cache, error_pages = set()):
        error_lookup = set()
        self.root.build_error_lookup(error_lookup, error_pages)
        out = [u'<div class="menu">'
               u'<h1>Gliederung</h1>'
               u'<ul>']
        get_image = file_cache.get_image
        for child in self.root.children:
            child.render(page, out, get_image, error_lookup)
        out.append(u"</ul>"
                   u"</div>")
        return u"".join(out)

    def getNextPage(self, page):
        try:
            return self.pages[page][1]
        except KeyError:
            return None

    def getTreeItem(self, item_name):
        stack = [self.root]
        while stack:
            item = stack.pop()
            if item.name == item_name:
                return item
            if item.children:
                stack.extend(item.children)
        return None

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
